#!/usr/bin/env python3
# coding=utf-8

# This file is part of PROMISE.
#
# 	PROMISE is free software: you can redistribute it and/or modify it
# 	under the terms of the GNU Lesser General Public License as
# 	published by the Free Software Foundation, either version 3 of the
# 	License, or (at your option) any later version.
#
# 	PROMISE is distributed in the hope that it will be useful, but WITHOUT
# 	ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# 	or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# 	Public License for more details.
#
# 	You should have received a copy of the GNU Lesser General Public
# 	License along with PROMISE. If not, see
# 	<http://www.gnu.org/licenses/>.
#
# Promise v1 was written by Romain Picot
# Promise v2 has been written from v1 by Thibault Hilaire and Sara Hoseininasab
# 	Sorbonne Université
# 	LIP6 (Computing Science Laboratory)
# 	Paris, France
# Contact: thibault.hilaire@lip6.fr
#
#
# File: runPromise.py
# Date: April 2020
#
# 	contain the entry function, called to run Promise
#
# 	© Thibault HILAIRE, April 2020


"""
\U0001f918 Promise v2 \U0001f918

Usage:
	runPromise -h | --help
	runPromise (hsd|hs|sd) [options]

Options:
  -h --help                     Show this screen.
  --conf CONF_FILE              get the configuration file [default: promise.yml]
  --output OUTPUT               set the path of the output (where the result files are put)
  --verbosity VERBOSITY         set the verbosity (betwen 0  and 4 for very low level debug) [default: 1]
  --log LOGFILE                 set the log file (no log file if this is not defined)
  --verbosityLog VERBOSITY      set the verbosity of the log file
  --debug                       put intermediate files into `debug/` (and `compileErrors/` for compilation errrors) and display the execution trace when an error comes
  --run RUN                     file to be run
  --compile COMMAND             command to compile the code
  --files FILES                 list of files to be examined by Promise (by default, all the .cc files)
  --nbDigits DIGITS             general required number of digits
  --path PATH                   set the path of the project (by default, the current path)
  --pause                       do pause between steps
  --noParsing                   do not parse the C file (__PROMISE__ are replaced and that's all)
  --alias ALIAS                 allow aliases (examples "g++=g++-14") [default:""]
  hsd                           Half/Single/Double mixed-precision
  hs                            Half/Single mixed-precision
  sd                            Single/Double mixed-precision

  
"""

from os.path import join
import sys
import docopt
from collections import Counter

from .promise import Promise
from .utils import parseOptions, PromiseError, getYMLOptions, Timing, pause, commaAnd
from .logger import PrLogger
from .prfile import PrFile


def runPromise(argv=None):
	"""This function is registered (in setup.py) as an entry_point
	argv is used for the unit tests"""

	# reset the logger and get a new instance
	logger = PrLogger()
	displayTrace = False


	# types handle by Promise
	typeNames = {'h': 'Half', 's': 'Single', 'd': 'Double', 'q': 'Quad'}
	types = {'h': 'half_float::half', 's': 'float', 'd': 'double', 'q': 'float128'}


	try:
		# get the options from command line and yml file
		args = docopt.docopt(__doc__, argv=sys.argv[1:] if argv is None else argv)      # parse the command line
		displayTrace = args['--debug']
		logger.configureLogger(args)                                                   # configure the logger
		options = getYMLOptions(args)                                           # get the options from the yml file
		logger.message("\U0001f918 Promise v2 \U0001f918")
		method, path, files, run, nbDigits, compileLines, outputPath, typeCustom, alias = parseOptions(options)    # parse the options
		PrFile.setCustom(typeCustom)
		compileErrorPath = join(path, 'compileErrors') if args['--debug'] else None
		tempPath = join(path, 'debug') if args['--debug'] else None


		
		# run with timing
		with Timing() as timing:
			# create Promise object
			pr = Promise(path, files, run, nbDigits, compileLines, parsing=not args['--noParsing'], alias=alias)

			# display general infos
			logger.info("We are working with %d file%s and %d different types" %
			            (pr.nbFiles, ('' if pr.nbFiles < 2 else 's'), pr.nbVariables))
			logger.info(pr.expectation())

			# debug the files
			if args['--debug']:
				pr.exportParsedFiles(tempPath)


			print("**** test - tempPath ", tempPath)
			# get the cadna reference
			highest = types['q'] if 'q' in method else types['d']
			pr.changeSameType(highest)
			logger.step("Get a reference result with Cadna (%s)" % highest)
			pr.compileAndRun(tempPath, cadna=True)
			if args['--pause']:
				pause()

			# try with the highest precision
			logger.step("Check with highest format (%s)" % typeNames[method[-1]])
			pr.changeSameType(types[method[-1]])
			
			if not pr.compileAndRun(tempPath):
				pr.changeSameType(highest)
				raise PromiseError("You should lower your expectation, it doesn't work with " + typeNames[method[-1]])
			
			if args['--pause']:
				pause()

			# do the Delta-Debug passes ('s','d' and then 'h','s' when method is 'hsd' for example)
			for lo, hi in reversed(list(zip(method, method[1:]))):
				logger.step("Delta-Debug %s/%s" % (typeNames[lo], typeNames[hi]))
				res = pr.runDeltaDebug(types[lo], types[hi], tempPath, args['--pause'], compileErrorPath)
				# stop if the DeltaDebug is not successful
				if not res:
					break

		# export the output
		if argv is None:
			pr.exportFinalResult(outputPath)


	except PromiseError as e:
		logger.error(e, exc_info=displayTrace)

	except docopt.DocoptExit:
		logger.error(__doc__)

	else:
		if timing:
			# display the number of each type
			count = Counter(pr.typesDict.values())  # count the nb of each type (result is a dictionary type:nb)
			li = ["%dx %s" % (v, k) for k, v in count.items()]
			logger.message("The final result contains %s.", commaAnd(li))
			logger.debug("Final types:\n" + pr.strResult())

			# display the stats
			logger.message("It tooks %.2fs", timing.timing)
			logger.message("\U0001F449 %d compilations (%d failed) for %.2fs", *pr.compilations)
			logger.message("\U0001F449 %d executions   (%d failed) for %.2fs", *pr.executions)

			logger.reset()
			return pr.setPerType()

	# reset the handlers, in case of running runPromise several times (otherwise, the log files are still open)
	logger.reset()


def runPromise_custom(argv=None):
	"""This function is registered (in setup.py) as an entry_point
	argv is used for the unit tests"""

	# reset the logger and get a new instance
	logger = PrLogger()
	displayTrace = False


	# types handle by Promise
	typeNames = {'h': 'Half', 's': 'Single', 'd': 'Double', 'q': 'Quad'}
	types = {'h': 'half_float::half', 's': 'float', 'd': 'double', 'q': 'float128'}


	try:
		# get the options from command line and yml file
		args = docopt.docopt(__doc__, argv=sys.argv[1:] if argv is None else argv)      # parse the command line
		displayTrace = args['--debug']
		logger.configureLogger(args)                                                   # configure the logger
		options = getYMLOptions(args)                                           # get the options from the yml file
		logger.message("\U0001f918 Promise v2 \U0001f918")
		method, path, files, run, nbDigits, compileLines, outputPath, typeCustom, alias = parseOptions(options)    # parse the options
		PrFile.setCustom(typeCustom)
		compileErrorPath = join(path, 'compileErrors') if args['--debug'] else None
		tempPath = join(path, 'debug') if args['--debug'] else None


		
		# run with timing
		with Timing() as timing:
			# create Promise object
			pr = Promise(path, files, run, nbDigits, compileLines, parsing=not args['--noParsing'], alias=alias)

			# display general infos
			logger.info("We are working with %d file%s and %d different types" %
			            (pr.nbFiles, ('' if pr.nbFiles < 2 else 's'), pr.nbVariables))
			logger.info(pr.expectation())

			# debug the files
			if args['--debug']:
				pr.exportParsedFiles(tempPath)


			print("**** test - tempPath ", tempPath)
			# get the cadna reference
			highest = types['q'] if 'q' in method else types['d']
			pr.changeSameType(highest)
			logger.step("Get a reference result with Cadna (%s)" % highest)
			pr.compileAndRun(tempPath, cadna=True)
			if args['--pause']:
				pause()

			# try with the highest precision
			logger.step("Check with highest format (%s)" % typeNames[method[-1]])
			pr.changeSameType(types[method[-1]])
			
			if not pr.compileAndRun(tempPath):
				pr.changeSameType(highest)
				raise PromiseError("You should lower your expectation, it doesn't work with " + typeNames[method[-1]])
			
			if args['--pause']:
				pause()

			# do the Delta-Debug passes ('s','d' and then 'h','s' when method is 'hsd' for example)
			for lo, hi in reversed(list(zip(method, method[1:]))):
				logger.step("Delta-Debug %s/%s" % (typeNames[lo], typeNames[hi]))
				res = pr.runDeltaDebug(types[lo], types[hi], tempPath, args['--pause'], compileErrorPath)
				# stop if the DeltaDebug is not successful
				if not res:
					break

		# export the output
		if argv is None:
			pr.exportFinalResult(outputPath)


	except PromiseError as e:
		logger.error(e, exc_info=displayTrace)

	except docopt.DocoptExit:
		logger.error(__doc__)

	else:
		if timing:
			# display the number of each type
			count = Counter(pr.typesDict.values())  # count the nb of each type (result is a dictionary type:nb)
			li = ["%dx %s" % (v, k) for k, v in count.items()]
			logger.message("The final result contains %s.", commaAnd(li))
			logger.debug("Final types:\n" + pr.strResult())

			# display the stats
			logger.message("It tooks %.2fs", timing.timing)
			logger.message("\U0001F449 %d compilations (%d failed) for %.2fs", *pr.compilations)
			logger.message("\U0001F449 %d executions   (%d failed) for %.2fs", *pr.executions)

			logger.reset()
			return pr.setPerType()

	# reset the handlers, in case of running runPromise several times (otherwise, the log files are still open)
	logger.reset()



if __name__ == "__main__":
	runPromise()



# this file was originally written by Zeller for Delta Debug
# see the original credtis (modifications logged at the end)

# $Id: DD.py,v 1.2 2001/11/05 19:53:33 zeller Exp $
# Enhanced Delta Debugging class
# Copyright (c) 1999, 2000, 2001 Andreas Zeller.

# This module (written in Python) implements the base delta debugging
# algorithms and is at the core of all our experiments.  This should
# easily run on any platform and any Python version since 1.6.
#
# To plug this into your system, all you have to do is to create a
# subclass with a dedicated `test()' method.  Basically, you would
# invoke the DD test case minimization algorithm (= the `ddmin()'
# method) with a list of characters; the `test()' method would combine
# them to a document and run the test.  This should be easy to realize
# and give you some good starting results; the file includes a simple
# sample application.
#
# This file is in the public domain; feel free to copy, modify, use
# and distribute this software as you wish - with one exception.
# Passau University has filed a patent for the use of delta debugging
# on program states (A. Zeller: `Isolating cause-effect chains',
# Saarland University, 2001).  The fact that this file is publicly
# available does not imply that I or anyone else grants you any rights
# related to this patent.
#
# The use of Delta Debugging to isolate failure-inducing code changes
# (A. Zeller: `Yesterday, my program worked', ESEC/FSE 1999) or to
# simplify failure-inducing input (R. Hildebrandt, A. Zeller:
# `Simplifying failure-inducing input', ISSTA 2000) is, as far as I
# know, not covered by any patent, nor will it ever be.  If you use
# this software in any way, I'd appreciate if you include a citation
# such as `This software uses the delta debugging algorithm as
# described in (insert one of the papers above)'.
#
# All about Delta Debugging is found at the delta debugging web site,
#
#               http://www.st.cs.uni-sb.de/dd/
#
# Happy debugging,
#
# Andreas Zeller


# Copyright 2016 Romain PICOT
#
# This file is part of PROMISE.
#
#    PROMISE is free software: you can redistribute it and/or modify it
#    under the terms of the GNU Lesser General Public License as
#    published by the Free Software Foundation, either version 3 of the
#    License, or (at your option) any later version.
#
#    PROMISE is distributed in the hope that it will be useful, but WITHOUT
#    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
#    or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
#    Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with PROMISE. If not, see
#    <http://www.gnu.org/licenses/>.
#


# Modification by Romain PICOT:
# Patch to use the Delta-Debug algorithm from
# "WHY PROGRAMS FAIL: A Guide to Systematic Debugging"

# Modification by Thibault HILAIRE (Sorbonne Université, LIP6, 2020)
# - port to Python3 (with 2to3), restore tabs, comply to PEP8
# - add (custom) logger
# - rewrite some parts in a more pythonic way


# TODO: remove unnecessary functions, tests, etc.
# TODO: remove the use of self.debug_xxxx (now dealed with the logger)

# get the logger
from cadnaPromise.logger import PrLogger
logger = PrLogger()


# Start with some helpers.
class OutcomeCache:
	"""This class holds test outcomes for configurations.  This avoids
	running the same test twice.

	The outcome cache is implemented as a tree.  Each node points
	to the outcome of the remaining list.

	Example: ([1, 2, 3], PASS), ([1, 2], FAIL), ([1, 4, 5], FAIL):

		 (2, FAIL)--(3, PASS)
		/
	(1, None)
		\
		 (4, None)--(5, FAIL)"""

	def __init__(self):
		self.tail = {}                  # Points to outcome of tail
		self.result = None              # Result so far

	def add(self, C, result):
		"""Add (C, RESULT) to the cache.  C must be a list of scalars."""
		cs = C[:]
		cs.sort()

		p = self
		for start in range(len(C)):
			if C[start] not in p.tail:
				p.tail[C[start]] = OutcomeCache()
			p = p.tail[C[start]]

		p.result = result

	def lookup(self, C):
		"""Return RESULT if (C, RESULT) is in the cache; None, otherwise."""
		p = self
		for start in range(len(C)):
			if C[start] not in p.tail:
				return None
			p = p.tail[C[start]]

		return p.result

	def lookup_superset(self, C, start=0):
		"""Return RESULT if there is some (C', RESULT) in the cache with
		C' being a superset of C or equal to C.  Otherwise, return None."""

		# FIXME: Make this non-recursive!
		if start >= len(C):
			if self.result:
				return self.result
			elif self.tail != {}:
				# Select some superset
				superset = self.tail[list(self.tail.keys())[0]]
				return superset.lookup_superset(C, start + 1)
			else:
				return None

		if C[start] in self.tail:
			return self.tail[C[start]].lookup_superset(C, start + 1)

		# Let K0 be the largest element in TAIL such that K0 <= C[START]
		k0 = None
		for k in list(self.tail.keys()):
			if (k0 is None or k > k0) and k <= C[start]:
				k0 = k

		if k0 is not None:
			return self.tail[k0].lookup_superset(C, start)

		return None

	def lookup_subset(self, C):
		"""Return RESULT if there is some (C', RESULT) in the cache with
		C' being a subset of C or equal to C.  Otherwise, return None."""
		p = self
		for start in range(len(C)):
			if C[start] in p.tail:
				p = p.tail[C[start]]

		return p.result


def oc_test():
	""""Test the outcome cache"""
	oc = OutcomeCache()
	assert oc.lookup([1, 2, 3]) is None
	oc.add([1, 2, 3], 4)
	assert oc.lookup([1, 2, 3]) == 4
	assert oc.lookup([1, 2, 3, 4]) is None

	assert oc.lookup([5, 6, 7]) is None
	oc.add([5, 6, 7], 8)
	assert oc.lookup([5, 6, 7]) == 8

	assert oc.lookup([]) is None
	oc.add([], 0)
	assert oc.lookup([]) == 0

	assert oc.lookup([1, 2]) is None
	oc.add([1, 2], 3)
	assert oc.lookup([1, 2]) == 3
	assert oc.lookup([1, 2, 3]) == 4

	assert oc.lookup_superset([1]) == 3 or oc.lookup_superset([1]) == 4
	assert oc.lookup_superset([1, 2]) == 3 or oc.lookup_superset([1, 2]) == 4
	assert oc.lookup_superset([5]) == 8
	assert oc.lookup_superset([5, 6]) == 8
	assert oc.lookup_superset([6, 7]) == 8
	assert oc.lookup_superset([7]) == 8
	assert oc.lookup_superset([]) is not None

	assert oc.lookup_superset([9]) is None
	assert oc.lookup_superset([7, 9]) is None
	assert oc.lookup_superset([-5, 1]) is None
	assert oc.lookup_superset([1, 2, 3, 9]) is None
	assert oc.lookup_superset([4, 5, 6, 7]) is None

	assert oc.lookup_subset([]) == 0
	assert oc.lookup_subset([1, 2, 3]) == 4
	assert oc.lookup_subset([1, 2, 3, 4]) == 4
	assert oc.lookup_subset([1, 3]) is None
	assert oc.lookup_subset([1, 2]) == 3

	assert oc.lookup_subset([-5, 1]) is None
	assert oc.lookup_subset([-5, 1, 2]) == 3
	assert oc.lookup_subset([-5]) == 0


# helpers
def listminus(C1, C2):
	"""Return a list of all elements of C1 that are not in C2."""
	s2 = {delta: 1 for delta in C2}
	C = [delta for delta in C1 if delta not in s2]
	return C


def listintersect(C1, C2):
	"""Return the common elements of C1 and C2."""
	s2 = {delta: 1 for delta in C2}
	C = [delta for delta in C1 if delta in s2]
	return C


def listunion(C1, C2):
	"""Return the union of C1 and C2."""
	s1 = {delta: 1 for delta in C1}
	C = C1[:] + [delta for delta in C2 if delta not in s1]
	return C


def listsubseteq(C1, C2):
	"""Return 1 if C1 is a subset or equal to C2."""
	s2 = {delta: 1 for delta in C2}

	for delta in C1:
		if delta not in s2:
			return 0

	return 1


# Main Delta Debugging algorithm.
class DD:
	"""Delta debugging base class.  To use this class for a particular
	setting, create a subclass with an overloaded `test()' method.

	Main entry points are:
	- `ddmin()' which computes a minimal failure-inducing configuration, and
	- `dd()' which computes a minimal failure-inducing difference.

	See also the usage sample at the end of this file.

	For further fine-tuning, you can implement an own `resolve()'
	method (tries to add or remove configuration elements in case of
	inconsistencies), or implement an own `split()' method, which
	allows you to split configurations according to your own
	criteria.

	The class includes other previous delta debugging alorithms,
	which are obsolete now; they are only included for comparison
	purposes."""

	# Test outcomes.
	PASS = "PASS"
	FAIL = "FAIL"
	UNRESOLVED = "UNRESOLVED"

	# Resolving directions.
	ADD = "ADD"                        # Add deltas to resolve
	REMOVE = "REMOVE"                        # Remove deltas to resolve

	# Debugging output (set to 1 to enable)
	debug_test = 0
	debug_dd = 0
	debug_split = 0
	debug_resolve = 0

	def __init__(self):
		self.__resolving = 0
		self.__last_reported_length = 0
		self.monotony = 0
		self.outcome_cache = OutcomeCache()
		self.cache_outcomes = 1
		self.minimize = 1
		self.maximize = 1
		self.assume_axioms_hold = 1

	# Output
	def coerce(self, C):
		"""Return the configuration C as a compact string"""
		# Default: use printable representation
		return repr(C)

	def pretty(self, C):
		"""Like coerce(), but sort beforehand"""
		sorted_c = C[:]
		sorted_c.sort()
		return self.coerce(sorted_c)

	# Testing
	def test(self, C):
		"""Test the configuration C.  Return PASS, FAIL, or UNRESOLVED"""
		C.sort()

		# If we had this test before, return its result
		if self.cache_outcomes:
			cached_result = self.outcome_cache.lookup(C)
			if cached_result is not None:
				return cached_result

		if self.monotony:
			# Check whether we had a passing superset of this test before
			cached_result = self.outcome_cache.lookup_superset(C)
			if cached_result == self.PASS:
				return self.PASS

			cached_result = self.outcome_cache.lookup_subset(C)
			if cached_result == self.FAIL:
				return self.FAIL

		if self.debug_test:
			logger.lowdebug("test(" + self.coerce(C) + ")...")

		outcome = self._test(C)

		if self.debug_test:
			logger.lowdebug("test(" + self.coerce(C) + ") = " + repr(outcome))

		if self.cache_outcomes:
			self.outcome_cache.add(C, outcome)

		return outcome

	def _test(self, C):
		"""Stub to overload in subclasses"""
		return self.UNRESOLVED                # Placeholder

	# Splitting

	def split(self, C, n):
		"""Split C into [C_1, C_2, ..., C_n]."""
		if self.debug_split:
			logger.lowdebug("split(" + self.coerce(C) + ", " + repr(n) + ")...")

		outcome = self._split(C, n)

		if self.debug_split:
			logger.lowdebug("split(" + self.coerce(C) + ", " + repr(n) + ") = " + repr(outcome))

		return outcome

	def _split(self, C, n):
		"""Stub to overload in subclasses"""
		subsets = []
		start = 0
		for i in range(n):
			subset = C[start:start + (len(C) - start) // (n - i)]
			subsets.append(subset)
			start = start + len(subset)
		return subsets


	# Resolving


	def resolve(self, csub, C, direction):
		"""If direction == ADD, resolve inconsistency by adding deltas
			 to CSUB.  Otherwise, resolve by removing deltas from CSUB."""

		if self.debug_resolve:
			logger.lowdebug("resolve(" + repr(csub) + ", " + self.coerce(C) + ", " + repr(direction) + ")...")

		outcome = self._resolve(csub, C, direction)

		if self.debug_resolve:
			logger.lowdebug("resolve(" + repr(csub) + ", " + self.coerce(C) + ", " + repr(direction) + ") = " + repr(outcome))

		return outcome


	def _resolve(self, csub, C, direction):
		"""Stub to overload in subclasses."""
		# By default, no way to resolve
		return None


	# Test with fixes


	def test_and_resolve(self, csub, r, C, direction):
		"""Repeat testing CSUB + R while unresolved."""

		initial_csub = csub[:]
		C2 = listunion(r, C)

		csubr = listunion(csub, r)
		t = self.test(csubr)

		# necessary to use more resolving mechanisms which can reverse each
		# other, can (but needn't) be used in subclasses
		self._resolve_type = 0

		while t == self.UNRESOLVED:
			self.__resolving = 1
			csubr = self.resolve(csubr, C, direction)

			if csubr is None:
				# Nothing left to resolve
				break
			else:
				if len(csubr) >= len(C2):
					# Added everything: csub == c2. ("Upper" Baseline)
					# This has already been tested.
					csubr = None
					break

				if len(csubr) <= len(r):
					# Removed everything: csub == r. (Baseline)
					# This has already been tested.
					csubr = None
					break

				t = self.test(csubr)

		self.__resolving = 0
		if csubr is None:
			return self.UNRESOLVED, initial_csub

		assert t == self.PASS or t == self.FAIL
		csub = listminus(csubr, r)
		return t, csub

	# Inquiries
	def resolving(self):
		"""Return 1 while resolving."""
		return self.__resolving


	# Logging


	def report_progress(self, C, title):
		if len(C) != self.__last_reported_length:
			logger.lowdebug(title + ": " + repr(len(C)) + " deltas left")
			self.__last_reported_length = len(C)


	# Delta Debugging (old ESEC/FSE version)



	def test_mix(self, csub, C, direction):
		t = self.FAIL   # default value, never used
		if self.minimize:
			t, csub = self.test_and_resolve(csub, [], C, direction)
			if t == self.FAIL:
				return t, csub

		if self.maximize:
			csubbar = listminus(self.CC, csub)
			cbar = listminus(self.CC, C)
			if direction == self.ADD:
				directionbar = self.REMOVE
			else:
				directionbar = self.ADD

			(tbar, csubbar) = self.test_and_resolve(csubbar, [], cbar, directionbar)

			csub = listminus(self.CC, csubbar)

			if tbar == self.PASS:
				t = self.FAIL
			elif tbar == self.FAIL:
				t = self.PASS
			else:
				t = self.UNRESOLVED

		return t, csub


	# Delta Debugging (new ISSTA version)


	def ddgen(self, C, minimize, maximize):
		"""Return a 1-minimal failing subset of C"""

		self.minimize = minimize
		self.maximize = maximize

		n = 2
		self.CC = C

		if self.debug_dd:
			logger.lowdebug(("dd(" + self.pretty(C) + ", " + repr(n) + ")..."))

		outcome = self._dd(C, n)

		if self.debug_dd:
			logger.lowdebug(("dd(" + self.pretty(C) + ", " + repr(n) + ") = " + repr(outcome)))

		return outcome

	def _dd(self, C, n):
		"""Stub to overload in subclasses"""

		# assert self.test([]) == self.PASS

		run = 1
		cbar_offset = 0

		# We replace the tail recursion from the paper by a loop
		while 1:
			# tc = self.test(c)
			# assert tc == self.FAIL or tc == self.UNRESOLVED

			if n > len(C):
				# No further minimizing
				logger.lowdebug("dd: done")
				return C

			self.report_progress(C, "dd")

			cs = self.split(C, n)

			logger.debug("dd (run #" + repr(run) + "): trying " + "+".join(str(len(cc)) for cc in cs))

			c_failed = 0
			cbar_failed = 0

			next_c = C[:]
			next_n = n

			# Check subsets
			# for i in range(n):
			#     if self.debug_dd:
			#         logger.debug "dd: trying", self.pretty(cs[i])

			#     (t, cs[i]) = self.test_mix(cs[i], c, self.REMOVE)

			#     if t == self.FAIL:
			#         # Found
			#         if self.debug_dd:
			#             logger.debug "dd: found", len(cs[i]), "deltas:",
			#             logger.debug self.pretty(cs[i])

			#         c_failed = 1
			#         next_c = cs[i]
			#         next_n = 2
			#         cbar_offset = 0
			#         self.report_progress(next_c, "dd")
			#         break

			if not c_failed:
				# Check complements
				cbars = n * [self.UNRESOLVED]

				# logger.debug "cbar_offset =", cbar_offset

				for j in range(n):
					if cbar_offset == 0:
						i = (j + cbar_offset) % (n + 1)
					else:
						i = (j + cbar_offset) % n
					cbars[i] = listminus(C, cs[i])
					t, cbars[i] = self.test_mix(cbars[i], C, self.ADD)

					doubled = listintersect(cbars[i], cs[i])
					if doubled is not []:
						cs[i] = listminus(cs[i], doubled)

					if t == self.FAIL:
						if self.debug_dd:
							logger.lowdebug("dd: reduced to" + str(len(cbars[i])) + "deltas:" + self.pretty(cbars[i]))

						cbar_failed = 1
						next_c = listintersect(next_c, cbars[i])
						next_n = next_n - 1
						self.report_progress(next_c, "dd")

						# In next run, start removing the following subset
						cbar_offset = i
						break

			if not c_failed and not cbar_failed:
				if n >= len(C):
					# No further minimizing
					logger.lowdebug("dd: done")
					return C

				next_n = min(len(C), n * 2)
				logger.lowdebug("dd: increase granularity to" + str(next_n))
				cbar_offset = (cbar_offset * next_n) // n

			C = next_c
			n = next_n
			run = run + 1

	def ddmin(self, C):
		return self.ddgen(C, 1, 0)

	def ddmax(self, C):
		return self.ddgen(C, 0, 1)

	def ddmix(self, C):
		return self.ddgen(C, 1, 1)


	# General delta debugging (new TSE version)


	def dddiff(self, C):
		n = 2

		if self.debug_dd:
			logger.lowdebug(("dddiff(" + self.pretty(C) + ", " + repr(n) + ")..."))

		outcome = self._dddiff([], C, n)

		if self.debug_dd:
			logger.lowdebug(("dddiff(" + self.pretty(C) + ", " + repr(n) + ") = " + repr(outcome)))

		return outcome

	def _dddiff(self, C1, C2, n):
		run = 1
		cbar_offset = 0

		# We replace the tail recursion from the paper by a loop
		while 1:
			if self.debug_dd:
				logger.lowdebug("dd: c1 =" + self.pretty(C1))
				logger.lowdebug("dd: c2 =" + self.pretty(C2))

			if self.assume_axioms_hold:
				t1 = self.PASS
				t2 = self.FAIL
			else:
				t1 = self.test(C1)
				t2 = self.test(C2)

			assert t1 == self.PASS
			assert t2 == self.FAIL
			assert listsubseteq(C1, C2)

			C = listminus(C2, C1)

			if self.debug_dd:
				logger.lowdebug("dd: c2 - c1 =" + self.pretty(C))

			if n > len(C):
				# No further minimizing
				logger.lowdebug("dd: done")
				return C, C1, C2

			self.report_progress(C, "dd")

			cs = self.split(C, n)

			logger.debug("dd (run #" + repr(run) + "): trying" + "+".join(str(len(cc)) for cc in cs))

			progress = 0

			next_c1 = C1[:]
			next_c2 = C2[:]
			next_n = n

			# Check subsets
			for j in range(n):
				i = (j + cbar_offset) % n

				if self.debug_dd:
					logger.lowdebug("dd: trying" + self.pretty(cs[i]))

				(t, csub) = self.test_and_resolve(cs[i], C1, C, self.REMOVE)
				csub = listunion(C1, csub)

				if t == self.FAIL and t1 == self.PASS:
					# Found
					progress = 1
					next_c2 = csub
					next_n = 2
					cbar_offset = 0

					if self.debug_dd:
						logger.lowdebug("dd: reduce c2 to" + str(len(next_c2)) + "deltas:" + self.pretty(next_c2))
					break

				if t == self.PASS and t2 == self.FAIL:
					# Reduce to complement
					progress = 1
					next_c1 = csub
					next_n = max(next_n - 1, 2)
					cbar_offset = i

					if self.debug_dd:
						logger.lowdebug("dd: increase c1 to" + str(len(next_c1)) + "deltas:" + self.pretty(next_c1))
					break


				csub = listminus(C, cs[i])
				(t, csub) = self.test_and_resolve(csub, C1, C, self.ADD)
				csub = listunion(C1, csub)

				if t == self.PASS and t2 == self.FAIL:
					# Found
					progress = 1
					next_c1 = csub
					next_n = 2
					cbar_offset = 0

					if self.debug_dd:
						logger.lowdebug("dd: increase c1 to" + str(len(next_c1)) + "deltas:" + self.pretty(next_c1))
					break

				if t == self.FAIL and t1 == self.PASS:
					# Increase
					progress = 1
					next_c2 = csub
					next_n = max(next_n - 1, 2)
					cbar_offset = i

					if self.debug_dd:
						logger.lowdebug("dd: reduce c2 to" + str(len(next_c2)) + "deltas:" + self.pretty(next_c2))
					break

			if progress:
				self.report_progress(listminus(next_c2, next_c1), "dd")
			else:
				if n >= len(C):
					# No further minimizing
					logger.lowdebug("dd: done")
					return C, C1, C2

				next_n = min(len(C), n * 2)
				logger.lowdebug("dd: increase granularity to" + str(next_n))
				cbar_offset = (cbar_offset * next_n) // n

			C1 = next_c1
			C2 = next_c2
			n = next_n
			run = run + 1

	def dd(self, C):
		return self.dddiff(C)           # Backwards compatibility




# coding=utf-8

"""This file is part of PROMISE.

	PROMISE is free software: you can redistribute it and/or modify it
	under the terms of the GNU Lesser General Public License as
	published by the Free Software Foundation, either version 3 of the
	License, or (at your option) any later version.

	PROMISE is distributed in the hope that it will be useful, but WITHOUT
	ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
	or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
	Public License for more details.

	You should have received a copy of the GNU Lesser General Public
	License along with PROMISE. If not, see
	<http://www.gnu.org/licenses/>.

Promise v1 was written by Romain Picot
Promise v2 has been written from v1 by Thibault Hilaire and Sara Hoseininasab
	Sorbonne Université
	LIP6 (Computing Science Laboratory)
	Paris, France
Contact: thibault.hilaire@lip6.fr


File: error.py
Date: April 2020

	custom empty classes for the errors and exceptions

	© Thibault HILAIRE, April 2020
"""


class PromiseError(Exception):
	"""Custom class to raise Promise Exception"""
	pass


class PromiseCompilationError(PromiseError):
	"""Custom class to raise Promise Exception when Compilation failed"""
	pass



from .promise import Promise 
import os
from os import environ

__version__ = '0.0.2'

curr_loc = os.path.dirname(os.path.realpath(__file__))
set_cadna_env = curr_loc+'/cadna/'
environ["CADNA_PATH"] = set_cadna_env



import os

def initCADNA():
    
    import subprocess
    import platform

    opt_system = platform.system()
    curr_loc = os.path.dirname(os.path.realpath(__file__))


    if not os.path.isfile('a.out'):
        if opt_system == 'Linux':
            os.chdir(curr_loc+r'/cadna')
            subprocess.call('pwd', shell=True)
            subprocess.call('bash run_linux.sh', shell=True)
        elif opt_system == 'Windows':
            os.chdir(curr_loc+r'/cadna')
            subprocess.call('pwd', shell=True)
            subprocess.call('bash run_mac.sh', shell=True)
        elif opt_system == 'posix' or opt_system == 'Darwin':
            os.chdir(curr_loc+r'/cadna')
            subprocess.call('pwd', shell=True)
            subprocess.call('bash run_win.sh', shell=True)

    return 1


# coding=utf-8

"""This file is part of PROMISE.

	PROMISE is free software: you can redistribute it and/or modify it
	under the terms of the GNU Lesser General Public License as
	published by the Free Software Foundation, either version 3 of the
	License, or (at your option) any later version.

	PROMISE is distributed in the hope that it will be useful, but WITHOUT
	ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
	or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
	Public License for more details.

	You should have received a copy of the GNU Lesser General Public
	License along with PROMISE. If not, see
	<http://www.gnu.org/licenses/>.

Promise v1 was written by Romain Picot
Promise v2 has been written from v1 by Thibault Hilaire and Sara Hoseininasab
	Sorbonne Université
	LIP6 (Computing Science Laboratory)
	Paris, France
Contact: thibault.hilaire@lip6.fr


File: utils.py
Date: April 2020

	Contains some useful functions (run some commands, parse the options, do some timing, etc.)

	© Thibault HILAIRE, April 2020
"""


from os import chdir
from os.path import exists, join
from yaml import load as loadyml, SafeLoader, YAMLError
from glob import glob
from subprocess import list2cmdline, STDOUT, check_output, CalledProcessError
from timeit import default_timer
from tqdm import tqdm
import os
from colorama import Fore

import shlex

from .errors import PromiseError
from .logger import PrLogger, getInt


logger = PrLogger()        # logger



def str2dict(s):
	"""transform a string in form 'param=value' to a dictionary {'param':'value'} """
	# see https://stackoverflow.com/questions/23228352/turn-key-value-string-into-a-dict
	return dict(x.split('=') for x in shlex.split(s))

def commaAnd(li):
	"""Returns a string where all the items of the list li are separated with a comma `,`
	except for the last one separated by `and` (and of cours there is no `and` if the
	list has only one item"""
	return ", ".join(li[0:-1]) + (' and ' if li[0:-1] else '') + li[-1]


def runCommand(cmd, alias={}, errorsAsDegug=False):
	"""run shell command and log it
	(and manage the errors)
	Returns True if everything was ok
	log the errors as debug message when errorsAsDebug is True"""
	# proceed to replacement in the command line according to the dictionary alias
	cmdline = list2cmdline(cmd)
	for k, v in alias.items():
		cmdline = cmdline.replace(k, v)
	# run the command
	try:
			logger.command('$ ' + cmdline)
			lines = check_output(cmdline, stderr=STDOUT, shell=True, universal_newlines=True,)
	except CalledProcessError as e:
		if errorsAsDegug:
			logger.command(e.output)
		else:
			logger.error(e.output)
		return False, ''
	# log the lines
	if lines:
		# display the line in `outputPromise` if it begins with `[PROMISE_`,  otherwise in `output`
		for li in lines.split('\n'):
			if li.startswith("[PROMISE_"):
				logger.outputPromise(li)
			else:
				logger.output(li)

	return True, lines.split('\n')


def cd(path):
	"""Change directory"""
	logger.command('$ cd '+path)
	chdir(path)




def getYMLOptions(args):
	"""get the options from the yaml file
	and merge them with those from the command line
	Returns a dictionary of the options {name:value}
	"""
	# read the config file (`promise.ymf` file  or the file given with '--conf' arg)
	ymlFileName = args['--conf'] if args['--conf'] else 'promise.yml'
	try:
		with open(ymlFileName, 'r') as f:
			dataset = loadyml(f.read(), SafeLoader)
	except IOError as err:
		if args['--conf']:
			logger.error("The file %s is empty or doesn't exist", ymlFileName)
			raise PromiseError(err) from err
	except YAMLError as err:
		logger.error("Cannot parse the %s file", ymlFileName)
		raise PromiseError(err) from err
	#
	conf = {'--'+key: True if value is None else value for key, value in dataset.items()}
	# add a file handler to the logger if it's in conf but not in args
	if '--log' in conf and not args['--log']:
		# get the path
		YMLpath = conf['--path'] if '--path' in conf else ''
		path = args['--path'] if args['--path'] else YMLpath
		# get the verbosity for the lg
		if not args['--verbosityLog']:
			verbosityLog = getInt(logger, conf, 'verbosityLog', 1, 4, default=int(args['--verbosity']))
		else:
			verbosityLog = args['--verbosityLog']
		# add the file logger
		logger.addFileLogger(path, conf['--log'], verbosityLog)

	# merge the two
	return {str(key): args.get(key) or conf.get(key) for key in set(conf) | set(args)}


def parseOptions(options):
	"""parse the dictionary of options
	Returns:
		- method: (string) method name for the mixed-precision ('hs', 'sd' or 'hsd')
		- path: (string) the path of the project
		- files: (list of string) list of filenames
		- run: (string) file to run
		- nbDigits: (int) nb of digits
		- compileLines: (list of string) commands to compile the project
		- output: (string) folder where put the generated code
		- typeCustom: (dictionary) custom types (used with __PRC_xxxx__)
		- alias: (string)
		"""
	# get the path
	path = options['--path'] if options['--path'] else ''

	# get the output path
	output = options['--output'] if options['--output'] else 'result'

	# list of files to examin (given by --files, or the *.c files in the path)
	if options['--files']:
		files = [name.strip() for name in options['--files'].split(',')]
	else:
		files = glob('*.c')
	for f in files:
		if not exists(join(path, f)):
			logger.error("The file %s doesn't exist in %s (declared in `--files` option)", f, path)
			raise PromiseError

	# run file
	run = options['--run']
	if not run:
		logger.error('The executable name is missing (`--run` option)')
		raise PromiseError

	# nb of digits
	try:
		nbDigitsGen = int(options['--nbDigits'])
	except ValueError as err:
		logger.error('The number of digits is missing or incorrect (`--nbDigits` option)')
		raise PromiseError(err) from err
	if '--nbDigitsPerVariable' in options:
		try:
			nbDigitsPerVar = {var: int(d) for var, d in options['--nbDigitsPerVariable'].items()}
		except (ValueError, AttributeError) as err:
			logger.error('The option `nbDigitsPerVariable` is incorrect. It should be a yml dictionary {varName: nbDigits}')
			raise PromiseError() from err
	else:
		nbDigitsPerVar = {}

	# compile commands
	if options['--compile']:
		compileLines = options['--compile']
	else:
		logger.error("No compilation command is given (`--compile` option)")
		raise PromiseError

	# method
	method = [k for k, v in options.items() if v and k in ("hsd", "sd", "hs")][0]

	# alias
	alias = str2dict(options['--alias']) if options['--alias'] else {}

	return method, path, files, run, (nbDigitsGen, nbDigitsPerVar), compileLines, output, options.get('--custom', None), alias


def pause(status=None):
	"""Make a pause
	wait for the user to press a key"""
	if status is None:
		tqdm.write("Press any key to continue...")
	else:
		status.set_description_str("%sPress any key to continue..." % Fore.RESET)
		status.refresh()
	os.system("""bash -c 'read -s -n 1'""")


class Timing:
	"""Small context manager to measure the time used
	it also increased the list [nbOccurenceTotal, nbOccurenceFailed, time] that is used for the constructor
	To be used as:

	>>> executions = [0,0]
	>>> with Timing(executions):
	>>>     ...
	>>>     #code_we_want_to_measure_time()
	>>>     ...

	and the elements of the list executions will be taken update accordingly

	We use here timeit.default_timer for the time (and not perf_counter or process_time)
	See https://stackoverflow.com/questions/7370801/measure-time-elapsed-in-python/25823885 for example"""
	def __init__(self, nbTi=None):
		"""nbTi is a 2-element list [nbOccurenceTotal, nbOccurenceFailed, time]
		those elements will be increased accordingly"""
		self._nbTi = nbTi
		self._timer = 0         # we haven't start yet
		self._timing = 0       # we haven't measure anything

	@property
	def timing(self):
		"""Return the time passed in the context"""
		return self._timing

	def __enter__(self):
		"""Enter in the context"""
		self._timer = default_timer()
		return self     # used to get the timing, after the context

	def __exit__(self, ty, value, traceback):
		""""We quit the context"""
		self._timing = default_timer() - self._timer
		if self._nbTi:
			self._nbTi[2] += self._timing
			self._nbTi[0] += 1
			if ty:
				self._nbTi[1] += 1
